# HG changeset patch
# User Vedran Vidovic <vvidovic@croz.net>
# Date 1652869454 -7200
#      Wed May 18 12:24:14 2022 +0200
# Node ID 42dbba2654acaeccc9a22a27c2c94eaaf754d680
# Parent  35afae4b3dffff6718c0cab3ceb16b9de207c20a
Option ssl_verify_partial_chain

Added configuration option `ssl_verify_partial_chain`
usable in modules:
- ngx_http_ssl_module
- ngx_mail_ssl_module
- ngx_stream_ssl_module

It sets the `X509_V_FLAG_PARTIAL_CHAIN` flag in the
openssl and enables us to add the intermediate certificates
to the list of trusted CA certificates for our clients, without
setting the root certificate of the chain.

Enables us to trust only the specific certificate without
widening the trust to it's parent certificate(s).

Example excerpt from the Nginx `http` / `server` configuration:
```
        ssl_client_certificate certs/my-intermediate-ca.pem;
        ssl_verify_client         on;
        ssl_verify_partial_chain  on;
```

diff -r 35afae4b3dff -r 42dbba2654ac contrib/vim/syntax/nginx.vim
--- a/contrib/vim/syntax/nginx.vim	Fri Apr 29 17:38:01 2022 +0400
+++ b/contrib/vim/syntax/nginx.vim	Wed May 18 12:24:14 2022 +0200
@@ -639,6 +639,7 @@
 syn keyword ngxDirective contained ssl_trusted_certificate
 syn keyword ngxDirective contained ssl_verify_client
 syn keyword ngxDirective contained ssl_verify_depth
+syn keyword ngxDirective contained ssl_verify_partial_chain
 syn keyword ngxDirective contained starttls
 syn keyword ngxDirective contained state
 syn keyword ngxDirective contained status
diff -r 35afae4b3dff -r 42dbba2654ac src/event/ngx_event_openssl.c
--- a/src/event/ngx_event_openssl.c	Fri Apr 29 17:38:01 2022 +0400
+++ b/src/event/ngx_event_openssl.c	Wed May 18 12:24:14 2022 +0200
@@ -866,7 +866,7 @@
 
 ngx_int_t
 ngx_ssl_client_certificate(ngx_conf_t *cf, ngx_ssl_t *ssl, ngx_str_t *cert,
-    ngx_int_t depth)
+    ngx_int_t depth, ngx_flag_t partial_chain)
 {
     STACK_OF(X509_NAME)  *list;
 
@@ -874,6 +874,25 @@
 
     SSL_CTX_set_verify_depth(ssl->ctx, depth);
 
+    if (partial_chain == 1) {
+      X509_VERIFY_PARAM *param = X509_VERIFY_PARAM_new();;
+      if (param) {
+        X509_VERIFY_PARAM_set_flags(param, X509_V_FLAG_PARTIAL_CHAIN);
+        if (SSL_CTX_set1_param(ssl->ctx, param) == 0) {
+          ngx_ssl_error(NGX_LOG_EMERG, ssl->log, 0,
+                        "SSL_CTX_set1_param() for flag "
+                        "X509_V_FLAG_PARTIAL_CHAIN failed.");
+          return NGX_ERROR;
+        }
+        X509_VERIFY_PARAM_free(param);
+
+      } else {
+        ngx_ssl_error(NGX_LOG_EMERG, ssl->log, 0,
+                      "X509_VERIFY_PARAM_new() failed.");
+        return NGX_ERROR;
+      }
+    }
+
     if (cert->len == 0) {
         return NGX_OK;
     }
diff -r 35afae4b3dff -r 42dbba2654ac src/event/ngx_event_openssl.h
--- a/src/event/ngx_event_openssl.h	Fri Apr 29 17:38:01 2022 +0400
+++ b/src/event/ngx_event_openssl.h	Wed May 18 12:24:14 2022 +0200
@@ -192,7 +192,7 @@
 ngx_int_t ngx_ssl_ciphers(ngx_conf_t *cf, ngx_ssl_t *ssl, ngx_str_t *ciphers,
     ngx_uint_t prefer_server_ciphers);
 ngx_int_t ngx_ssl_client_certificate(ngx_conf_t *cf, ngx_ssl_t *ssl,
-    ngx_str_t *cert, ngx_int_t depth);
+    ngx_str_t *cert, ngx_int_t depth, ngx_flag_t partial_chain);
 ngx_int_t ngx_ssl_trusted_certificate(ngx_conf_t *cf, ngx_ssl_t *ssl,
     ngx_str_t *cert, ngx_int_t depth);
 ngx_int_t ngx_ssl_crl(ngx_conf_t *cf, ngx_ssl_t *ssl, ngx_str_t *crl);
diff -r 35afae4b3dff -r 42dbba2654ac src/http/modules/ngx_http_ssl_module.c
--- a/src/http/modules/ngx_http_ssl_module.c	Fri Apr 29 17:38:01 2022 +0400
+++ b/src/http/modules/ngx_http_ssl_module.c	Wed May 18 12:24:14 2022 +0200
@@ -170,6 +170,13 @@
       offsetof(ngx_http_ssl_srv_conf_t, verify_depth),
       NULL },
 
+    { ngx_string("ssl_verify_partial_chain"),
+      NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_CONF_FLAG,
+      ngx_conf_set_flag_slot,
+      NGX_HTTP_SRV_CONF_OFFSET,
+      offsetof(ngx_http_ssl_srv_conf_t, verify_partial_chain),
+      NULL },
+
     { ngx_string("ssl_client_certificate"),
       NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_CONF_TAKE1,
       ngx_conf_set_str_slot,
@@ -586,6 +593,7 @@
     sscf->buffer_size = NGX_CONF_UNSET_SIZE;
     sscf->verify = NGX_CONF_UNSET_UINT;
     sscf->verify_depth = NGX_CONF_UNSET_UINT;
+    sscf->verify_partial_chain = NGX_CONF_UNSET;
     sscf->certificates = NGX_CONF_UNSET_PTR;
     sscf->certificate_keys = NGX_CONF_UNSET_PTR;
     sscf->passwords = NGX_CONF_UNSET_PTR;
@@ -804,7 +812,8 @@
 
         if (ngx_ssl_client_certificate(cf, &conf->ssl,
                                        &conf->client_certificate,
-                                       conf->verify_depth)
+                                       conf->verify_depth,
+                                       conf->verify_partial_chain)
             != NGX_OK)
         {
             return NGX_CONF_ERROR;
diff -r 35afae4b3dff -r 42dbba2654ac src/http/modules/ngx_http_ssl_module.h
--- a/src/http/modules/ngx_http_ssl_module.h	Fri Apr 29 17:38:01 2022 +0400
+++ b/src/http/modules/ngx_http_ssl_module.h	Wed May 18 12:24:14 2022 +0200
@@ -27,6 +27,7 @@
 
     ngx_uint_t                      verify;
     ngx_uint_t                      verify_depth;
+    ngx_flag_t                      verify_partial_chain;
 
     size_t                          buffer_size;
 
diff -r 35afae4b3dff -r 42dbba2654ac src/mail/ngx_mail_ssl_module.c
--- a/src/mail/ngx_mail_ssl_module.c	Fri Apr 29 17:38:01 2022 +0400
+++ b/src/mail/ngx_mail_ssl_module.c	Wed May 18 12:24:14 2022 +0200
@@ -188,6 +188,13 @@
       offsetof(ngx_mail_ssl_conf_t, verify_depth),
       NULL },
 
+    { ngx_string("ssl_verify_partial_chain"),
+      NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_CONF_FLAG,
+      ngx_conf_set_flag_slot,
+      NGX_HTTP_SRV_CONF_OFFSET,
+      offsetof(ngx_http_ssl_srv_conf_t, verify_partial_chain),
+      NULL },
+
     { ngx_string("ssl_client_certificate"),
       NGX_MAIL_MAIN_CONF|NGX_MAIL_SRV_CONF|NGX_CONF_TAKE1,
       ngx_conf_set_str_slot,
@@ -331,6 +338,7 @@
     scf->prefer_server_ciphers = NGX_CONF_UNSET;
     scf->verify = NGX_CONF_UNSET_UINT;
     scf->verify_depth = NGX_CONF_UNSET_UINT;
+    scf->verify_partial_chain = NGX_CONF_UNSET;
     scf->builtin_session_cache = NGX_CONF_UNSET;
     scf->session_timeout = NGX_CONF_UNSET;
     scf->session_tickets = NGX_CONF_UNSET;
diff -r 35afae4b3dff -r 42dbba2654ac src/mail/ngx_mail_ssl_module.h
--- a/src/mail/ngx_mail_ssl_module.h	Fri Apr 29 17:38:01 2022 +0400
+++ b/src/mail/ngx_mail_ssl_module.h	Wed May 18 12:24:14 2022 +0200
@@ -31,6 +31,7 @@
 
     ngx_uint_t       verify;
     ngx_uint_t       verify_depth;
+    ngx_flag_t       verify_partial_chain;
 
     ssize_t          builtin_session_cache;
 
diff -r 35afae4b3dff -r 42dbba2654ac src/stream/ngx_stream_ssl_module.c
--- a/src/stream/ngx_stream_ssl_module.c	Fri Apr 29 17:38:01 2022 +0400
+++ b/src/stream/ngx_stream_ssl_module.c	Wed May 18 12:24:14 2022 +0200
@@ -156,6 +156,13 @@
       offsetof(ngx_stream_ssl_conf_t, verify_depth),
       NULL },
 
+    { ngx_string("ssl_verify_partial_chain"),
+      NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_CONF_FLAG,
+      ngx_conf_set_flag_slot,
+      NGX_HTTP_SRV_CONF_OFFSET,
+      offsetof(ngx_http_ssl_srv_conf_t, verify_partial_chain),
+      NULL },
+
     { ngx_string("ssl_client_certificate"),
       NGX_STREAM_MAIN_CONF|NGX_STREAM_SRV_CONF|NGX_CONF_TAKE1,
       ngx_conf_set_str_slot,
@@ -676,6 +683,7 @@
     scf->prefer_server_ciphers = NGX_CONF_UNSET;
     scf->verify = NGX_CONF_UNSET_UINT;
     scf->verify_depth = NGX_CONF_UNSET_UINT;
+    scf->verify_partial_chain = NGX_CONF_UNSET;
     scf->builtin_session_cache = NGX_CONF_UNSET;
     scf->session_timeout = NGX_CONF_UNSET;
     scf->session_tickets = NGX_CONF_UNSET;
diff -r 35afae4b3dff -r 42dbba2654ac src/stream/ngx_stream_ssl_module.h
--- a/src/stream/ngx_stream_ssl_module.h	Fri Apr 29 17:38:01 2022 +0400
+++ b/src/stream/ngx_stream_ssl_module.h	Wed May 18 12:24:14 2022 +0200
@@ -26,6 +26,7 @@
 
     ngx_uint_t       verify;
     ngx_uint_t       verify_depth;
+    ngx_flag_t       verify_partial_chain;
 
     ssize_t          builtin_session_cache;
 
